import sys
from typing import Callable, Dict, List, Optional, Union
from collections.abc import Iterable

# Import literal type for different versions of Python.
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from .chars import CHARS

# Colors class.
from .color import color as c

# Styling function.
from .utils import style

# Quick styles.
s  : Callable[[str], str] = style('s')
d  : Callable[[str], str] = style('d')
i  : Callable[[str], str] = style('i')
u  : Callable[[str], str] = style('u')
k  : Callable[[str], str] = style('k')
h  : Callable[[str], str] = style('h')
x  : Callable[[str], str] = style('x')
du : Callable[[str], str] = style('du')
rev: Callable[[str], str] = style('rev')

# Quick colors.
r  : Callable[[str], str] = style('r')
g  : Callable[[str], str] = style('g')
y  : Callable[[str], str] = style('y')
b  : Callable[[str], str] = style('b')
m  : Callable[[str], str] = style('m')
cy : Callable[[str], str] = style('c')
w  : Callable[[str], str] = style('w')

def available() -> List[str]:
    """ Returns the available styles.

    Returns:
        List[str]: The available styles.
    """

    # Get the available styles.
    ints: Dict[int, str] = {v: '' for v in set(CHARS.values())}
    for sty in CHARS.keys():
        i: int = CHARS[sty]
        if i in ints.keys() and not ints[i]:
            ints[i] = sty
    
    return list(ints.values())

def color(r: int, g: int, b: int, type: Literal['fg', 'bg'] = 'fg') -> str:
    """ Returns the ANSI escape sequence for the given RGB color.
    
    Args:
        r (int): The red value.
        g (int): The green value.
        b (int): The blue value.
        type (str): The type of color. (defaults to 'fg')

    Returns:
        str: The ANSI escape sequence.
    """

    # Check if the type is valid.
    if type not in ['fg', 'bg']:
        raise ValueError("Type must be 'fg' (foreground) or 'bg' (background).")

    # Check if the RGB values are all integers.
    if not all(isinstance(v, int) for v in [r, g, b]):
        raise TypeError("RGB values must be integers.")
    
    # Check if the RGB values are in the range 0-255.
    if not all(0 <= v <= 255 for v in [r, g, b]):
        raise ValueError("RGB values must be in the range 0-255.")

    # Set the type.
    type_: Literal['38', '48'] = '38' if type == 'fg' else '48'
    
    # Generate the ANSI escape sequence.
    code: str = f'CODE{type_};2;{r};{g};{b}'
    
    return code

def prints(
    *values: object,
    s: Optional[Union[str, List[str], Callable[[str], str]]] = None,
    **kwargs: object
) -> None:
    """ Prints the given values with the given style. Sends the output to Python's print function.

    Args:
        *values (object): The values to print.
        s (Optional[Union[str, List[str], Callable[[str], str]]]): The style to apply to the text. (defaults to None)
        **kwargs (object): The keyword arguments to pass to the print function.
    """

    # Set the style to empty function.
    style_: Callable[[str], str] = style()

    if not s:
        # No style was given.
        pass
    elif isinstance(s, str):
        # The style is a single string.
        style_ = style(s)
    elif isinstance(s, Iterable):
        # The style is an array.
        style_ = style(*s)
    elif callable(s):
        # The style is a function.
        style_ = s
    else:
        # Invalid style.
        raise TypeError('Style must be an array or a function.')

    # Print the text.
    print(*[style_(str(v)) for v in values], kwargs)


from typing import Callable, List, Union
import warnings

from .chars import CHARS

# Returns the ANSI escape sequence for the given code.
sand: Callable[[object], str] = lambda code: '\x1b[' + str(code) + 'm'

def apply(text: object, *codes: str) -> str:
    """ Applies the given styles to the text.

    Args:
        text (object): The text to style.
        *codes (str): The styles to apply to the text.
        p (bool): Whether to print the text. (defaults to False)

    Returns:
        str: The styled text if p is False.
    """

    # Check if codes only contains strings.
    if not all(isinstance(c, str) for c in codes):
        raise TypeError("Codes must be strings.")

    # Combine the styles and text.
    style_: str = ''.join(codes)
    reset: str = sand(CHARS['reset'])
    text_: str = style_ + str(text) + reset

    return text_

def style(*styles: Union[int, Callable[[str], str], str]) -> Callable[[str], str]:
    """ Returns a function that applies the given styles to the text.

    Args:
        *styles (Union[int, Callable[[str], str], str]): The styles to apply to the text.

    Returns:
        Callable[[str], str]: The function that applies the styles to the text.
    """

    if not styles:
        # No styles were given.
        return lambda text: text
    
    # Convert the styles to a list.
    styles_: List[Union[int, Callable[[str], str], str]] = list(styles)

    # Get the ANSI escape sequences for the given styles.
    codes: List[str] = []
    for sty in styles_:
        
        # The integer value.
        if isinstance(sty, int):
            if sty not in CHARS.values():
                warnings.warn("Style with escape code " + str(sty) + " may not be recognized.", UserWarning)
            codes.append(str(sty))

        elif callable(sty):
            # Remove sty from the styles.
            styles_.remove(sty)

            # Return the style with the function.
            return lambda text: sty(style(*styles_)(text))
        
        elif not isinstance(sty, str):
            raise TypeError("Style " + sty + " must be an integer, a string, or a callable.")

        # Predefined code.
        elif sty.startswith('CODE'):
            codes.append(sty[4:])
        
        # User-input style.
        else:
            if sty not in CHARS.keys():
                raise ValueError(f"Style '{sty}' is not recognized.")
            codes.append(str(CHARS[str(sty).lower()]))
    
    codes = [sand(c) for c in codes]
    
    # Return the lambda function responsible for applying those styles.
    return lambda text: apply(text, *codes)


from setuptools import setup, find_packages
import sys

# Determine extra dependencies for specific Python versions
extras = []
if sys.version_info < (3, 8):
    extras.append('typing_extensions')
    
with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='AnsiLib',
    version='1.0.3',
    author='Bora Boyacıoğlu',
    author_email='boyacioglu20@itu.edu.tr',
    description='A library for handling ANSI codes',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/boraboyacioglu-itu/ansilib',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    license='MIT License',
    python_requires='>=3.5',
    install_requires=extras,
)


from typing import Callable
from .utils import style

class color:
    # Text colors.
    class t:
        k: Callable[[str], str] = style('kt0')
        r: Callable[[str], str] = style('rt0')
        g: Callable[[str], str] = style('gt0')
        y: Callable[[str], str] = style('yt0')
        b: Callable[[str], str] = style('bt0')
        m: Callable[[str], str] = style('mt0')
        c: Callable[[str], str] = style('ct0')
        w: Callable[[str], str] = style('wt0')
        d: Callable[[str], str] = style('dt0')

    # Background colors.
    class b:
        k: Callable[[str], str] = style('kb0')
        r: Callable[[str], str] = style('rb0')
        g: Callable[[str], str] = style('gb0')
        y: Callable[[str], str] = style('yb0')
        b: Callable[[str], str] = style('bb0')
        m: Callable[[str], str] = style('mb0')
        c: Callable[[str], str] = style('cb0')
        w: Callable[[str], str] = style('wb0')
        d: Callable[[str], str] = style('db0')

    # Bright text colors.
    class t_:
        k: Callable[[str], str] = style('kt1')
        r: Callable[[str], str] = style('rt1')
        g: Callable[[str], str] = style('gt1')
        y: Callable[[str], str] = style('yt1')
        b: Callable[[str], str] = style('bt1')
        m: Callable[[str], str] = style('mt1')
        c: Callable[[str], str] = style('ct1')
        w: Callable[[str], str] = style('wt1')

    # Bright background colors.
    class b_:
        k: Callable[[str], str] = style('kb1')
        r: Callable[[str], str] = style('rb1')
        g: Callable[[str], str] = style('gb1')
        y: Callable[[str], str] = style('yb1')
        b: Callable[[str], str] = style('bb1')
        m: Callable[[str], str] = style('mb1')
        c: Callable[[str], str] = style('cb1')
        w: Callable[[str], str] = style('wb1')